
import numpy as np
from .utils import _row_f, _check_dissimilarity_matrix

def upper_bound_samples(D: np.ndarray, kappa: int = 2) -> np.ndarray:
    """
    ...
    """

    _check_dissimilarity_matrix(D=D)
    
    # Remove diagonal from distance matrix and then sort
    D_hat = np.sort(D[~np.eye(D.shape[0],dtype=bool)].reshape(D.shape[0],-1))

    n = D_hat.shape[0]
    if n < 4:
        raise ValueError("Matrix must be at least of size 4x4.")
    
    if kappa < 1 or kappa > n - 2:
        raise ValueError("The parameter kappa is out of range.")
    
    # Compute bounds
    bounds = np.apply_along_axis(lambda row: _row_f(row, kappa=kappa, n=n), axis=1, arr=D_hat)

    return bounds

def upper_bound(D: np.ndarray, kappa: int = 2) -> float:
    """
    ...
    """

    point_bounds = upper_bound_samples(D=D, kappa=kappa)

    return np.mean(point_bounds)





"""
In this example, we show how to lower the upper bound assuming the restriction
that no cluster is smaller than kappa, where kappa > 2. 
"""

import numpy as np
from silhouette_upper_bound import upper_bound
from sklearn.cluster import KMeans
from sklearn.datasets import make_blobs
from sklearn.metrics import silhouette_score, pairwise_distances

def main():
    # Generate synthetic data
    X, _ = make_blobs(n_samples=3000, n_features=32, centers=7, random_state=42)
    D = pairwise_distances(X)
    
    # Cluster with KMeans
    kmeans = KMeans(n_clusters=7, random_state=42, n_init="auto")
    labels = kmeans.fit_predict(X)
    
    # Overall silhouette score
    score = silhouette_score(X, labels)
    print(f"KMeans Silhouette score: {score:.3f}")

    print(f"Silhouette upper bound: {upper_bound(D):.3f}")

    # Upper bound with kappa > 2
    min_cluster_size = np.bincount(labels).min() 
    print(f"Silhouette upper bound (given that we don't allow cluster sizes smaller than {min_cluster_size}): {upper_bound(D, kappa=min_cluster_size):.3f}")

if __name__ == "__main__":
    main()



import numpy as np
import pytest
from silhouette_upper_bound import upper_bound, upper_bound_samples
from sklearn.cluster import KMeans
from sklearn.datasets import make_blobs
from sklearn.metrics import silhouette_score, silhouette_samples, pairwise_distances

def test_basic():

    d1 = np.array([[0, 1, 1, 1, 1], 
                   [1, 0, 1, 1, 1], 
                   [1, 1, 0, 1, 1], 
                   [1, 1, 1, 0, 1], 
                   [1, 1, 1, 1, 0]])

    assert upper_bound(d1) == 0.0 

    d2 = np.array([[0, 1, 5, 5, 5],  
                   [1, 0, 5, 5, 5],  
                   [5, 5, 0, 1, 1],  
                   [5, 5, 1, 0, 1],  
                   [5, 5, 1, 1, 0]]) 

    assert upper_bound(d2) == 1 - 1 / 5

    d3 = np.array([[0, 1, 6, 6.5, 7, 8],  
                   [1, 0, 5, 7.5, 6.88, 8],  
                   [6, 5, 0, 0.5, 0.78, 1.1],  
                   [6.5, 7.5, 0.5, 0, 0.5, 1],  
                   [7, 6.88, 0.78, 0.5, 0, 1],
                   [8, 8, 1.1, 1, 1, 0]]) 
    
    d3f = np.array([4 * 1 / (6 + 6.5 + 7 + 8), 
                    4 * 1 / (5 + 7.5 + 6.88 + 8), 
                    (2 / 3) * (0.5 + 0.78 + 1.1) / (6 + 5), 
                    (2 / 3) * (0.5 + 0.5 + 1) / (6.5 + 7.5), 
                    (2 / 3) * (0.78 + 0.5 + 1) / (7 + 6.88),
                    (2 / 3) * (1.1 + 1 + 1) / (8 + 8)])
    
    assert np.abs(upper_bound(d3) - np.mean(1 - d3f)) < 1e-15


@pytest.mark.parametrize("n_samples", [100, 200, 300, 400, 500])
@pytest.mark.parametrize("n_features", [10, 15, 20])
@pytest.mark.parametrize("centers", [3, 6, 9])
@pytest.mark.parametrize("cluster_std", [1.0, 2.0, 3.0])
def test_blobs(n_samples, n_features, centers, cluster_std):
    X, _ = make_blobs(
        n_samples=n_samples,
        n_features=n_features,
        centers=centers,
        cluster_std=cluster_std,
        random_state=42,
    )

    D = pairwise_distances(X)

    # KMeans clustering
    model = KMeans(n_clusters=centers, random_state=42, n_init="auto")
    labels = model.fit_predict(X)
    score = silhouette_score(X, labels)

    # Test standard upper bound
    ub = upper_bound(D)

    assert 0 <= ub and ub <= 1
    assert ub - score >= - 1e-15

    # Test each sample 
    score_samples = silhouette_samples(X, labels)
    ub_samples = upper_bound_samples(D)
    assert np.all(0 <= ub_samples) and np.all(ub_samples <= 1)
    assert np.all(ub_samples - score_samples >= - 1e-15)

    # Upper bound with kappa > 2
    min_cluster_size = np.bincount(labels).min() 

    if min_cluster_size > 2:

        ub_kappa = upper_bound(D, kappa=min_cluster_size)

        assert 0 <= ub_kappa and ub_kappa <= 1
        assert ub_kappa - score >= - 1e-15
    
        # Test each sample 
        ub_kappa_samples = upper_bound_samples(D, kappa=min_cluster_size)
        assert np.all(0 <= ub_kappa_samples) and np.all(ub_kappa_samples <= 1)
        assert np.all(ub_kappa_samples - score_samples >= - 1e-15)




import numpy as np


def _check_dissimilarity_matrix(D: np.ndarray, tol: float = 1e-15):

    # Check that D is a valid dissimilarity matrix
    if D.ndim != 2 or D.shape[0] != D.shape[1]:
        raise ValueError("Matrix must be square.")
    
    if not np.all(D >= -tol):
        raise ValueError("Matrix must be non-negative.")
    
    if not np.allclose(np.diag(D), 0, atol=tol):
        raise ValueError("Matrix must have zero diagonal.")
    
    if not np.allclose(D, D.T, atol=tol):
        raise ValueError("Matrix must be symmetric.")


def _row_f(row: np.ndarray, kappa: int, n: int) -> float:
    
    x = np.sum(row[:kappa - 1])
        
    y = np.sum(row[kappa - 1:])

    q = (x / (kappa - 1)) / (y / (n - kappa))

    for delta in range(kappa + 1, n - kappa + 1):

        d_to_move = row[delta - 2]

        x += d_to_move
        y -= d_to_move
        
        q_candidate = (x / (delta - 1)) / (y / (n - delta))

        if q_candidate < q:
            q = q_candidate
    
    return 1 - q 
